/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"


#include <cmath>
#include <QDebug>

#include "AggregatedTreeItem.h"
#include "CubeMetric.h"
#include "CubeProxy.h"
#include "CubeRegion.h"
#include "CubeIdIndexMap.h"

#include "MetricTree.h"
#include "Globals.h"
#include "TreeView.h"
#include "AggregatedTreeItem.h"

using namespace std;
using namespace cube;
using namespace cubegui;

MetricTree::MetricTree()
{
    treeType = METRICTREE;
}

QString
MetricTree::getLabel()
{
    if ( !cube )
    {
        return tr( "not yet initialized" );
    }
    // todo proxy return ( cube->getAttribute( "metrictree title" ) != "" ) ? cube->getAttribute( "metrictree title" ).c_str() : "Metric tree";
    return tr( "Metric tree" );
}

TreeItem*
MetricTree::createTree()
{
    TreeItem* top = new TreeItem( this, QString(), METRICITEM, 0 );

    vector<cube::Metric*> metrics = cube->getRootMetrics();

    if ( Globals::optionIsSet( ExpertMode ) ) // also show ghost metrics
    {
        vector<cube::Metric*> ghost = cube->getGhostMetrics();
        metrics.insert( metrics.end(), ghost.begin(), ghost.end() );
    }
    createItems<cube::Metric>( top, metrics, METRICITEM );

    return top;
}

/** return false, if the item is invalid and should not be added to the tree */
bool
MetricTree::itemIsValid( cube::Vertex* vertex )
{
    // if the metric item is VOID, it will not inserted into the tree
    cube::Metric* metric     = ( cube::Metric* )vertex;
    bool          voidValued = false;
    while ( true )
    {
        if ( std::strcmp( metric->get_val().c_str(), "VOID" ) == 0 )
        {
            voidValued = true;
            break;
        }
        metric = metric->get_parent();
        if ( metric == NULL )
        {
            break;
        }
    }
    return !voidValued;
}

QString
MetricTree::getItemName( cube::Vertex* vertex )
{
    return QString( ( ( cube::Metric* )vertex )->get_disp_name().c_str() );
}

double
MetricTree::getMaxValue( const TreeItem* item ) const
{
    return maxValueIndex.size() > 0  ? maxValues.at( maxValueIndex.value( item ) ) : std::nan( "" );
}

/**
 * @brief fills the hash derivedDependencies, which contains for each derived metric item a list of metric items on which it depends
 * see MetricTree::getDependencyHash()
 */
void
MetricTree::initializeFullDependencyHash()
{
    derivedDependencies.clear();
    foreach( TreeItem * item, unavailableItems )
    {
        delete item;
    }
    unavailableItems.clear();

    QList<TreeItem*>          derivedMetrics; // list with all derived metrics from metric tree
    QHash<Metric*, TreeItem*> metricHash;     // metric object -> metric tree item
    foreach( TreeItem * item, treeItems )
    {
        cube::Metric* metric = static_cast<cube::Metric*> ( item->getCubeObject() );
        metricHash.insert( metric, item );
        if ( item->isDerivedMetric() )
        {
            derivedMetrics.append( item );
        }
    }
    if ( derivedMetrics.isEmpty() )
    {
        return;
    }

    foreach( TreeItem * item, derivedMetrics )
    {
        std::vector<cube::Metric*> dependentMetrics;
        cube::Metric*              metric = static_cast<cube::Metric*> ( item->getCubeObject() );
        metric->fillReqMetrics( dependentMetrics );
        QList<TreeItem*> dependentItems;
        foreach( Metric * metric, dependentMetrics )
        {
            if ( !metricHash.contains( metric ) ) // invisible ghost metric or inactive metric
            {
                TreeItem* item = new TreeItem( this, QString( metric->get_disp_name().c_str() ) + " " + metric->get_uniq_name().c_str(), METRICITEM, metric );
                metricHash.insert( metric, item );
                unavailableItems.append( item );
            }
            TreeItem* item = metricHash.value( metric );
            if ( !dependentItems.contains( item ) )
            {
                dependentItems.append( item );
            }
        }
        derivedDependencies.insert( item, dependentItems );
    }
}

/**
 * @brief MetricTree::getDependencyHash resolves dependencies of the derived metric items which should be calculated next in parallel.
 * Each derived metric depends on other metrics. If multiple metrics that depend on the same metric are calculated in parallel, the
 * execution time will increase because of locking.
 * The dependencies of all derived metric items items are stored in the hash derivedDependencies (seeMetricTree::initializeFullDependencyHash()).
 * If only a part of the tree is calculated, the missing dependencies have to be added to the list of items to calculate.
 * A derived metric depends on other metrics. If multiple metrics that depend on the same (not calculated) metric are calculated in parallel, the
 * execution time will increase because of locking.
 * A metric may depend on a metric that isn't calculated in the following step (= it isn't part of metrics or derived metrics). In that
 * case, this dependency will be added to the result hash.
 * @param metrics list of metrics, which will be calculated
 * @param derivedMetrics list of derived metrics, which will be calculated
 * Based on the result of resolveDependencies, a hash with current dependencies is generated. All
 * @return metric->dependencies hash
 */
QHash< TreeItem*, QList<TreeItem*> >
MetricTree::getDependencyHash( const QList<TreeItem*>& metrics, const QList<TreeItem*>& derivedMetrics )
{
    QHash< TreeItem*, QList<TreeItem*> > hash;                                                                 // return value
    QSet<TreeItem*>                      allItemSet = QSet<TreeItem*>::fromList( derivedDependencies.keys() ); // all derived metrics
    QSet<TreeItem*>                      derivedSet = QSet<TreeItem*>::fromList( derivedMetrics );             // all derived metrics to calculate
    QSet<TreeItem*>                      metricSet  = QSet<TreeItem*>::fromList( metrics );                    // all non-derived metrics to calculate

    QHash< TreeItem*, TreeItem* > unresolvedDeps;                                                              // metrics that are not part of the metric tree

    // could be called once after tree is created/modified, but cost of this method call is very low
    initializeFullDependencyHash();

    QList<TreeItem*> metricsToCalculate = derivedMetrics;
    while ( !metricsToCalculate.isEmpty() )
    {
        TreeItem*        item         = metricsToCalculate.takeLast();
        QList<TreeItem*> dependencies = derivedDependencies[ item ];
        foreach( TreeItem * dep, derivedDependencies[ item ] )
        {
            if ( dep->isCalculated() ) // dependency is already calculated -> remove it
            {
                dependencies.removeOne( dep );
                continue;
            }
            if ( !metricSet.contains( dep ) && !derivedSet.contains( dep ) )
            {
                // add dependent (currently unvisible) item to list of items to calculate
                if ( allItemSet.contains( dep ) )
                {
                    metricsToCalculate.append( dep );
                    derivedSet.insert( dep );
                }
                else // unresolveable dependency -> remove dependency of current item and add dependency to current item to all following occurences
                {
                    dependencies.removeOne( dep );
                    if ( !unresolvedDeps.contains( dep ) ) // first unresolvable dependency
                    {
                        unresolvedDeps.insert( dep, item );
                    }
                    else // following unresolvable dependency -> refer to first
                    {
                        dependencies.append( unresolvedDeps.value( dep ) );
                    }
                }
            }
        }
        hash.insert( item, dependencies );
    }

    return hash;
}

QVector<Task*>
MetricTree::setBasicValues( const QList<Tree*>& leftTrees,
                            const QList<Tree*>& rightTrees, const QList<TreeItem*>& items )
{
    QVector<Task*> workerData;

    if ( items.size() == 0 )
    {
        return workerData;
    }

    QList<TreeItem*> itemsToCalculate = items;
    QList<TreeItem*> derivedMetrics;
    foreach( TreeItem * child, itemsToCalculate )
    {
        if ( child->isSelected() ) // calculate selected items first
        {
            itemsToCalculate.removeOne( child );
            itemsToCalculate.prepend( child );
        }
        if ( child->isDerivedMetric() ) // derived metrics are computationally expensive -> calculate later
        {
            itemsToCalculate.removeOne( child );
            derivedMetrics.append( child );
        }
    }

    // Set up tree pointers
    Tree* callTree          = 0;
    Tree* callFlat          = 0;
    Tree* systemTree        = 0;
    bool  rightOfCallDim    = false;
    bool  rightOfSystemTree = false;


    // search in the left trees
    foreach( Tree * item, leftTrees )
    {
        if ( item->getTreeType() == CALLTREE )
        {
            callTree       = item;
            rightOfCallDim = true;
        }
        else if ( item->getTreeType() == CALLFLAT )
        {
            callFlat       = item;
            rightOfCallDim = true;
        }
        else if ( item->getTreeType() == SYSTEMTREE )
        {
            systemTree        = item;
            rightOfSystemTree = true;
        }
    }

    // search in the right trees
    if ( !systemTree || ( !callTree && !callFlat ) )
    {
        foreach( Tree * item, rightTrees )
        {
            if ( item->getTreeType() == CALLTREE )
            {
                callTree = item;
            }
            else if ( item->getTreeType() == CALLFLAT )
            {
                callFlat = item;
            }
            else if ( item->getTreeType() == SYSTEMTREE )
            {
                systemTree = item;
            }
        }
    }

    assert( callTree || callFlat );
    assert( systemTree );

    // Set up selection for system tree
    // if system tree is right, leave sysres_selection empty
    list_of_sysresources sysres_selection;
    if ( rightOfSystemTree )  // metric tree is on the right
    {
        QList< TreeItem* > selected_systemnodes = systemTree->getSelectionList();
        foreach( TreeItem * item, selected_systemnodes )
        {
            sysres_pair sp;
            sp.first  = static_cast<cube::Sysres*> ( item->getCubeObject() );
            sp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            sysres_selection.push_back( sp );
        }
    }
    else
    {
        sysres_selection = CubeProxy::ALL_SYSTEMNODES;
    }

    if ( callTree )
    {
        // Set up selection for call tree
        // if call tree is right, leave sysres_selection empty
        list_of_cnodes cnode_selection;

        if ( rightOfCallDim )  // metric tree is on the right
        {
            QList< TreeItem* > selected_cnodes = callTree->getSelectionList();
            foreach( TreeItem * item, selected_cnodes )
            {
                cnode_pair cp;
                cp.first  = static_cast<cube::Cnode*> ( item->getCubeObject() );
                cp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                cnode_selection.push_back( cp );
            }
        }
        else
        {
            cnode_selection = CubeProxy::ALL_CNODES;
        }

        int depth = 0; // 0 = first level; todo optimize proxy
        // request bulk data from CubeProxy
        // calculate simple metrics
        foreach( TreeItem * child, itemsToCalculate )
        {
            cube::Metric*   metric = static_cast<cube::Metric*> ( child->getCubeObject() );
            MetricTreeData* data   = new MetricTreeData( child,
                                                         cnode_selection,
                                                         sysres_selection,
                                                         metric,
                                                         depth,
                                                         QList<TreeItem*>() );

            workerData.append( data );
        }

        // calculate derived metrics
        QHash< TreeItem*, QList<TreeItem*> > dependencies = getDependencyHash( itemsToCalculate, derivedMetrics );
        foreach( TreeItem * child, dependencies.keys() )
        {
            cube::Metric*   metric = static_cast<cube::Metric*> ( child->getCubeObject() );
            MetricTreeData* data   = new MetricTreeData( child,
                                                         cnode_selection,
                                                         sysres_selection,
                                                         metric,
                                                         depth,
                                                         dependencies.value( child ) );
            workerData.append( data );
        }
    }
    else if ( callFlat )
    {
        // Set up selection for call flat
        list_of_regions region_selection;

        // Marks that some regions got status "subroutines" and have t be reset after calculation
        bool unset_regions = false;

        if ( rightOfCallDim )  // metric tree is on the right
        {
            QList< TreeItem* > selected_regions = callFlat->getSelectionList();
            foreach( TreeItem * item, selected_regions )
            {
                region_pair rp;
                rp.first = static_cast<cube::Region*> ( item->getCubeObject() );
                if ( rp.first == NULL ) // subroutines
                {
                    unset_regions = true;
                    cube::Region* reg = static_cast<cube::Region*> ( item->getParent()->getCubeObject() );
                    reg->set_as_subroutines();
                    rp.first = reg;
                }
                rp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                region_selection.push_back( rp );
            }
        }
        else
        {
            // Flat tree selection is the inclusive root node(s)
            QList< TreeItem* > selected_regions = callFlat->getTop().getChildren();
            TreeItem*          item             = selected_regions.at( 0 );
            region_pair        rp;
            rp.first  = static_cast<cube::Region*> ( item->getCubeObject() );
            rp.second = cube::CUBE_CALCULATE_INCLUSIVE;
            region_selection.push_back( rp );
        }

        workerData.append( new MetricTreeFlatData( this, region_selection, sysres_selection ) );

        if ( unset_regions )
        {
            for ( list_of_regions::iterator riter = region_selection.begin(); riter != region_selection.end(); ++riter )
            {
                ( *riter ).first->unset_as_subroutines();
            }
        }
    } // callflat
    return workerData;
}


/** used to calculate the data of the complete tree in a thread */
void
MetricTreeFlatData::calculate()
{
    // request bulk data from CubeProxy
    vector< Value* > inclusive_values;
    vector< Value* > exclusive_values;

    // get values for full tree
    tree->getCube()->getMetricTreeValues( region_selection,
                                          sysres_selection,
                                          inclusive_values,
                                          exclusive_values );

    foreach( TreeItem * child, tree->getTop().getChildren() )
    {
        // Assign values to all toplevel trees
        tree->assignValues( child, inclusive_values, exclusive_values );
    }
}

/** used to calculate the data of a tree item in a thread */
void
MetricTreeData::calculate()
{
    QElapsedTimer timer;
    timer.start();
    MetricTree*      tree = static_cast<MetricTree*> ( item->getTree() );
    cube::CubeProxy* cube = tree->getCube();

    std::vector< Value* > inclusive_values;
    std::vector< Value* > exclusive_values;
    cube::IdIndexMap      indexMap;

    std::vector< Value* > empty;
    if ( item->isExpanded() )
    {
        cube->getMetricSubtreeValues( cnode_selection,
                                      sysres_selection,
                                      *root_metric,
                                      depth,
                                      indexMap,
                                      cube::CUBE_PROXY_VALUES_IGNORE,
                                      &exclusive_values );
        tree->assignValues( item, empty, exclusive_values, item->getDepth() + depth, &indexMap );
    }
    else
    {
        cube->getMetricSubtreeValues( cnode_selection,
                                      sysres_selection,
                                      *root_metric,
                                      depth,
                                      indexMap,
                                      &inclusive_values,
                                      cube::CUBE_PROXY_VALUES_IGNORE );
        tree->assignValues( item, inclusive_values, empty, item->getDepth() + depth, &indexMap );
    }
}

/** returns true, if all dependent items have already been calculated */
bool
MetricTreeData::isReady()
{
    foreach( TreeItem * dep, dependentItems )
    {
        if ( !dep->isCalculated() )
        {
            return false;
        }
    }
    return true;
}



// -----------------------------------------------------------------
// ---------------- calculation functions --------------------------
// -----------------------------------------------------------------

void
MetricTree::computeMaxValues()
{
    double roundThreshold = Globals::getRoundThreshold( FORMAT_TREES );
    double value;
    double maxValue = 100.0;

    maxValues.clear();
    maxValueIndex.clear();

    // for metric trees we compute the max values for each root's subtree
    QList<TreeItem*> subTree;

    for ( unsigned i = 0; i < ( unsigned )top->getChildren().size(); i++ )
    {
        subTree.clear();
        subTree.append( top->getChildren().at( i ) );
        maxValue = top->getChildren().at( i )->totalValue;

        for ( int j = 0; j < subTree.size(); j++ )
        {
            maxValueIndex.insert( subTree[ j ], i );
            foreach( TreeItem * child, subTree[ j ]->getChildren() )
            {
                subTree.append( child );
            }
            getMax( subTree[ j ], value );
            // we save max absolute value (for the case of differences of cube)
            if ( fabs( maxValue ) < fabs( value ) )
            {
                maxValue = value;
            }
        }

        if ( fabs( maxValue ) <= fabs( roundThreshold ) )
        {
            maxValue = 0.0;
        }
        maxValues.append( maxValue );
    }
} // end of computeMaxValues()


TreeItem*
MetricTree::getTreeItem( std::string metricId )
{
    foreach( TreeItem * item, this->getItems() )
    {
        cube::Metric* metric = static_cast<cube::Metric*>( item->cubeObject );
        if ( metric->get_uniq_name() == metricId )
        {
            return item;
        }
    }
    return 0;
}





// return the name if the ith external metric item
//
double
MetricTree::getExternalReferenceValue( const QString& name )
{
    return externalMetricValue.value( name, 0.0 );
}


// compute the needed external values from the "cubeExternal"
// database for the external percentage value modus
//
void
MetricTree::computeExternalReferenceValues( cube::CubeProxy* cubeExternal )
{
    externalMetricValue.clear();

    const std::vector<cube::Metric*>& metricVec = cubeExternal->getMetrics();
    const std::vector<cube::Cnode*>&  callVec   = cubeExternal->getCnodes();
    const std::vector<cube::Thread*>& threadVec = cubeExternal->getLocations();

    // calculate inclusive values for all metric root items
    for ( int i = 0; i < ( int )metricVec.size(); i++ )
    {
        // take the next metric item
        cube::Metric* metric     = metricVec[ i ];
        QString       metricName = QString::fromStdString( metric->get_uniq_name() );

        // compute external metric values
        double totalValue = 0.0;
        for ( int j = 0; j < ( int )callVec.size(); j++ )
        {
            for ( int k = 0; k < ( int )threadVec.size(); k++ )
            {
                /// @fixme @todo Implement proxy computation for external reference values
                //totalValue += cube->get_sev( metric, callVec[ j ], threadVec[ k ] );
            }
        }
        externalMetricValue[ metricName ] = totalValue;
    }
}
