/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef METRICTREE_H
#define METRICTREE_H

#include "Tree.h"
#include <QHash>

namespace cubegui
{
class MetricTreeData;

class MetricTree : public Tree
{
public:
    MetricTree();

    virtual QString
    getLabel();

    TreeItem*
    getTreeItem( std::string metricId );

    /** returns the maximum value of the tree which contains the given item */
    virtual double
    getMaxValue( const TreeItem* item = 0 ) const;

    void
    computeExternalReferenceValues( cube::CubeProxy* cubeExternal );

    double
    getExternalReferenceValue( const QString& name );

protected:
    virtual TreeItem*
    createTree();

    virtual QString
    getItemName( cube::Vertex* vertex );

    virtual bool
    itemIsValid( cube::Vertex* vertex );

    virtual void
    computeMaxValues();

    /**
     * Set the basic inclusive and exclusive values in the full tree.
     * @param left  List of trees to the left of this tree
     * @param right List of trees to the right of this tree
     * @param itemsToCalculate, if list is not empty, only the given items are calculated, otherwise all visible items are calculated
     */
    virtual QVector<Task*>
    setBasicValues( const QList<Tree*>&     leftTrees,
                    const QList<Tree*>&     rightTrees,
                    const QList<TreeItem*>& itemsToCalculate = QList<TreeItem*>( ) );

private:
    // ---- calculation functions
    void
    computeValue( TreeItem* item );

    void
    computeValue( cube::list_of_cnodes& cv,
                  TreeItem*             item );
    void
    computeValue( cube::list_of_regions& rv,
                  TreeItem*              item );
    void
    computeValue( cube::list_of_sysresources& sv,
                  TreeItem*                   item );

    void
    computeValue( cube::list_of_regions&      rv,
                  cube::list_of_sysresources& sv,
                  TreeItem*                   item );
    void
    computeValue( cube::list_of_cnodes&       cv,
                  cube::list_of_sysresources& sv,
                  TreeItem*                   item );

    void
    initializeFullDependencyHash();

    QHash<TreeItem*, QList<TreeItem*> >
    getDependencyHash( const QList<TreeItem*>& metrics,
                       const QList<TreeItem*>& derivedMetrics );

    QHash< TreeItem*, QList<TreeItem*> > derivedDependencies; // see getDependencyHash()
    QList<TreeItem*>                     unavailableItems;    // see getDependencyHash()

    // value modus related attribute
    QHash<const TreeItem*, int> maxValueIndex;
    QList<double>               maxValues;

    // external metric items and their names for the external percentage value modus
    QList<TreeItem*>       externalMetricItems;
    QHash<QString, double> externalMetricValue;
    QList<std::string>     externalMetricNames;

    friend class MetricTreeFlatData;
    friend class MetricTreeData;
};

/** for use in QConcurrent */
class MetricTreeData : public Task
{
public:
    MetricTreeData( TreeItem* item_, const cube::list_of_cnodes& c, const cube::list_of_sysresources& s, cube::Metric* root, size_t depth_, QList<TreeItem*> deps ) :
        item( item_ ), cnode_selection( c ), sysres_selection( s ), root_metric( root ), depth( depth_ ), dependentItems( deps )
    {
    }

    virtual void
    calculate();

    virtual bool
    isReady();

private:
    TreeItem*                        item;
    const cube::list_of_cnodes       cnode_selection;
    const cube::list_of_sysresources sysres_selection;
    cube::Metric*                    root_metric;
    size_t                           depth;
    QList<TreeItem*>                 dependentItems; // items which have to be calculated before current item can be calculated
};

class MetricTreeFlatData : public Task
{
public:
    MetricTreeFlatData( MetricTree* t, const cube::list_of_regions& r, const cube::list_of_sysresources& s ) :
        tree( t ), region_selection( r ), sysres_selection( s )
    {
    }
    virtual void
    calculate();

private:
    MetricTree*                      tree;
    const cube::list_of_regions      region_selection;
    const cube::list_of_sysresources sysres_selection;
};
}

#endif // METRICTREE_H
