/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "POPCommunicationEfficiencyTest.h"


using namespace advisor;


POPCommunicationEfficiencyTest::POPCommunicationEfficiencyTest( cube::CubeProxy*      cube,
                                                                POPSerialisationTest* _pop_ser,
                                                                POPTransferTest*      _pop_transeff ) : PerformanceTest( cube ),
    pop_ser( _pop_ser ),
    pop_transeff( _pop_transeff )
{
    scout_cubex = scout_metrics_available( cube );
    setName( tr( "    Communication Efficiency" ).toUtf8().data() );
    setWeight( 1 );   // need to be adjusted
    pop_commeff = NULL;

    if ( !scout_cubex )
    {
        pop_commeff = cube->getMetric( "comm_eff" );
        if ( pop_commeff == NULL )
        {
            adjustForTest( cube );
        }
        pop_commeff = cube->getMetric( "comm_eff" );
        if ( pop_commeff == NULL )
        {
            setWeight( 0.2 );
            setValue( 0. );
            return;
        }
        cube::list_of_sysresources lsysres = getRootsOfSystemTree();
        cube::metric_pair          metric;
        metric.first  = pop_commeff;
        metric.second = cube::CUBE_CALCULATE_INCLUSIVE;
        lmetrics.push_back( metric );
        return;
    }
    setWeight( 0.2 );
    setValue( 0. );
}

void
POPCommunicationEfficiencyTest::calculateForScout()
{
    if ( pop_ser == NULL || pop_transeff == NULL )
    {
        return;
    }
    double pop_ser_value     = pop_ser->value();
    double pop_transef_value = pop_transeff->value();
    setValue( pop_ser_value *   pop_transef_value );
}


void
POPCommunicationEfficiencyTest::applyCnode( const cube::Cnode* cnode, const cube::CalculationFlavour cnf  )
{
    if ( !scout_cubex )
    {
        if ( pop_commeff == NULL )
        {
            return;
        }
        cube::value_container      inclusive_values;
        cube::value_container      exclusive_values;
        cube::list_of_sysresources lsysres; //  = getRootsOfSystemTree();
        cube::IdIndexMap           metric_id_indices;
        cube->getCallpathSubtreeValues( lmetrics,
                                        lsysres,
                                        *( const_cast<cube::Cnode*>( cnode ) ),
                                        0,
                                        metric_id_indices,
                                        &inclusive_values,
                                        &exclusive_values );
        double pop_commeff_value = ( cnf == cube::CUBE_CALCULATE_INCLUSIVE ) ?
                                   inclusive_values[ 0 ]->getDouble() :
                                   exclusive_values[ 0 ]->getDouble();
        setValue( pop_commeff_value );
    }
    else  // scout cubex
    {
        calculateForScout();
    }
}

void
POPCommunicationEfficiencyTest::applyCnode( const cube::list_of_cnodes& cnodes )
{
    if ( !scout_cubex )
    {
        if ( pop_commeff == NULL )
        {
            return;
        }
        cube::value_container      inclusive_values;
        cube::value_container      exclusive_values;
        cube::list_of_sysresources lsysres /* = getRootsOfSystemTree()*/;
        cube::IdIndexMap           metric_id_indices;

        cube->getMetricSubtreeValues( cnodes,
                                      lsysres,
                                      *pop_commeff,
                                      0,
                                      metric_id_indices,
                                      &inclusive_values,
                                      &exclusive_values );
        double pop_commeff_value = inclusive_values[ 0 ]->getDouble();
        setValue( pop_commeff_value  );
    }
    else
    {
        calculateForScout();
    }
}


const std::string&
POPCommunicationEfficiencyTest::getCommentText() const
{
    if ( pop_ser == NULL || pop_transeff == NULL )
    {
        return no_comment;
    }
    else
    {
        return no_comment;
    }
}


// ------ overview tests ---------

bool
POPCommunicationEfficiencyTest::isActive() const
{
    return ( pop_ser != NULL && pop_transeff != NULL )
           ||
           pop_commeff != NULL;
};

bool
POPCommunicationEfficiencyTest::isIssue() const
{
    return false;
};

void
POPCommunicationEfficiencyTest::adjustForTest( cube::CubeProxy* cube ) const
{
    add_comm_eff( cube );
}

// ----------------- private static calls ------------
void
POPCommunicationEfficiencyTest::add_comm_eff( cube::CubeProxy* cube ) const
{
    cube::Metric* _comp      = cube->getMetric( "comp" );
    cube::Metric* _execution = cube->getMetric( "execution" );
    if ( _comp == NULL )
    {
        add_comp_time( cube );
    }
    if ( _execution == NULL )
    {
        add_execution_time( cube );
    }


    cube::Metric* _met = cube->defineMetric(
        tr( "Communication efficiency per process" ).toUtf8().data(),
        "comm_eff",
        "DOUBLE",
        "",
        "",
        POP_COMM_EFF_METRIC_URL,
        tr( "Communication efficiency reflects the loss of efficiency by communication. It is computed as a ratio between maximum computation time and total runtime, i.e. ( comp / execution ). The value varies from 0 to 100%, where 100% shows that routine is a pure computation routine, whereas 0 shows that it is a MPI routine. Do not consider values in the flat view and the system tree." ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_POSTDERIVED,
        "{"
        " if( metric::fixed::mpi_indicator()>0)"
        "{"
        "      ${tmp}= metric::fixed::mpi_indicator(i) * metric::comp() / metric::execution();"
        "}"
        "else"
        "{"
        "      ${tmp} = metric::comp() / metric::execution();"
        "};"
        "return ${tmp};"
        "}",
        "",
        "",
        "",
        "max(arg1, arg2)",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}
