/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "POPIPCTest.h"


using namespace advisor;

POPIPCTest::POPIPCTest( cube::CubeProxy* cube ) : PerformanceTest( cube )
{
    setName( tr( "IPC" ).toUtf8().data() );
    setWeight( 1 );   // need to be adjusted
    pop_ipc = cube->getMetric( "ipc" );
    if ( pop_ipc == NULL )
    {
        adjustForTest( cube );
    }
    pop_ipc  = cube->getMetric( "ipc" );
    maxValue = 1.;
    if ( pop_ipc == NULL )
    {
        setWeight( 0.1 );
        setValue( 0. );
        return;
    }
    cube::list_of_sysresources lsysres = getRootsOfSystemTree();
    cube::metric_pair          metric;
    metric.first  = pop_ipc;
    metric.second = cube::CUBE_CALCULATE_INCLUSIVE;
    lmetrics.push_back( metric );
}


void
POPIPCTest::applyCnode( const cube::Cnode* cnode, const cube::CalculationFlavour cnf  )
{
    if ( pop_ipc == NULL )
    {
        return;
    }

    cube::value_container      inclusive_values;
    cube::value_container      exclusive_values;
    cube::list_of_sysresources lsysres;
    cube::IdIndexMap           metric_id_indices;
    cube->getCallpathSubtreeValues( lmetrics,
                                    lsysres,
                                    *( const_cast<cube::Cnode*>( cnode ) ),
                                    0,
                                    metric_id_indices,
                                    &inclusive_values,
                                    &exclusive_values );
    double pop_ipc_value = ( cnf == cube::CUBE_CALCULATE_INCLUSIVE ) ?
                           inclusive_values[ 0 ]->getDouble() :
                           exclusive_values[ 0 ]->getDouble();
    setValue( pop_ipc_value );
}

void
POPIPCTest::applyCnode( const cube::list_of_cnodes& cnodes )
{
    if ( pop_ipc == NULL )
    {
        return;
    }
    cube::value_container      inclusive_values;
    cube::value_container      exclusive_values;
    cube::list_of_sysresources lsysres;
    cube::IdIndexMap           metric_id_indices;
    cube->getMetricSubtreeValues( cnodes,
                                  lsysres,
                                  *pop_ipc,
                                  0,
                                  metric_id_indices,
                                  &inclusive_values,
                                  &exclusive_values );
    double pop_ipc_value = inclusive_values[ 0 ]->getDouble();
    setValue( pop_ipc_value );
}

const std::string&
POPIPCTest::getCommentText() const
{
    if ( pop_ipc == NULL )
    {
        return no_comment;
    }
    else
    {
        return no_comment;
    }
}


// ------ overview tests ---------

bool
POPIPCTest::isActive() const
{
    return pop_ipc != NULL;
};

bool
POPIPCTest::isIssue() const
{
    return false;
};


void
POPIPCTest::adjustForTest( cube::CubeProxy* cube ) const
{
    cube::Metric* _tot_ins_without_wait = cube->getMetric( "tot_ins_without_wait" );
    cube::Metric* _tot_cyc_without_wait = cube->getMetric( "tot_cyc_without_wait" );
    if ( _tot_ins_without_wait == NULL )
    {
        add_tot_ins_without_wait( cube );
    }
    if ( _tot_cyc_without_wait == NULL )
    {
        add_tot_cyc_without_wait( cube );
    }
    _tot_ins_without_wait = cube->getMetric( "tot_ins_without_wait" );
    _tot_cyc_without_wait = cube->getMetric( "tot_cyc_without_wait" );
    if ( _tot_ins_without_wait == NULL || _tot_cyc_without_wait == NULL )
    {
        return;
    }
    add_ipc( cube );
}



void
POPIPCTest::add_ipc( cube::CubeProxy* cube ) const
{
    cube::Metric* _met = cube->defineMetric(
        tr( "IPC" ).toUtf8().data(),
        "ipc",
        "DOUBLE",
        "",
        "",
        POP_IPC_METRIC_URL,
        tr( "Value of IPC (instructions per cycle) without busy-wait in MPI and OpenMP, computed as tot_ins_without_wait() / tot_cyc_without_wait()" ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_POSTDERIVED,
        "metric::tot_ins_without_wait() / metric::tot_cyc_without_wait()",
        "",
        "",
        "",
        "",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}



void
POPIPCTest::add_tot_ins_without_wait( cube::CubeProxy* cube ) const
{
    cube::Metric* PAPI_TOT_INS = cube->getMetric( "PAPI_TOT_INS" );
    if ( PAPI_TOT_INS == NULL )
    {
        return;
    }
    cube::Metric* _met = cube->defineMetric(
        tr( "PAPI_TOT_INS without busy-wait" ).toUtf8().data(),
        "tot_ins_without_wait",
        "DOUBLE",
        "",
        "",
        POP_IPC_METRIC_URL,
        tr( "Here is PAPI_TOT_INS without busy-wait in MPI and OpenMP." ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE,
        "${without_wait_state}[${calculation::callpath::id}] * metric::PAPI_TOT_INS()",
        "",
        "",
        "",
        "",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}



void
POPIPCTest::add_tot_cyc_without_wait( cube::CubeProxy* cube )  const
{
    cube::Metric* PAPI_TOT_CYC = cube->getMetric( "PAPI_TOT_CYC" );
    if ( PAPI_TOT_CYC == NULL )
    {
        return;
    }
    cube::Metric* _met = cube->defineMetric(
        tr( "PAPI_TOT_CYC without busy-wait" ).toUtf8().data(),
        "tot_cyc_without_wait",
        "DOUBLE",
        "",
        "",
        POP_IPC_METRIC_URL,
        tr( "Here is PAPI_TOT_CYC without busy-wait in MPI and OpenMP." ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE,
        "${without_wait_state}[${calculation::callpath::id}] * metric::PAPI_TOT_CYC()",
        "",
        "",
        "",
        "",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}
