/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "POPTransferTest.h"


using namespace advisor;

POPTransferTest::POPTransferTest( cube::CubeProxy* cube ) : PerformanceTest( cube )
{
    setName( tr( "         Transfer efficiency" ).toUtf8().data() );
    setWeight( 1 );   // need to be adjusted
    pop_transfer = cube->getMetric( "transf_eff" );
    if ( pop_transfer == NULL )
    {
        adjustForTest( cube );
    }
    pop_transfer = cube->getMetric( "transf_eff" );
    if ( pop_transfer == NULL )
    {
        setWeight( 0.1 );
        setValue( 0. );
        return;
    }
    cube::list_of_sysresources lsysres = getRootsOfSystemTree();
    cube::metric_pair          metric;
    metric.first  = pop_transfer;
    metric.second = cube::CUBE_CALCULATE_INCLUSIVE;
    lmetrics.push_back( metric );
}



void
POPTransferTest::applyCnode( const cube::Cnode* cnode, const cube::CalculationFlavour cnf  )
{
    if ( pop_transfer == NULL )
    {
        return;
    }
    cube::value_container      inclusive_values;
    cube::value_container      exclusive_values;
    cube::list_of_sysresources lsysres;
    cube::IdIndexMap           metric_id_indices;
    cube->getCallpathSubtreeValues( lmetrics,
                                    lsysres,
                                    *( const_cast<cube::Cnode*>( cnode ) ),
                                    0,
                                    metric_id_indices,
                                    &inclusive_values,
                                    &exclusive_values );
    double pop_transfer_value = ( cnf == cube::CUBE_CALCULATE_INCLUSIVE ) ?
                                inclusive_values[ 0 ]->getDouble() :
                                exclusive_values[ 0 ]->getDouble();
    setValue( pop_transfer_value );
}

void
POPTransferTest::applyCnode( const cube::list_of_cnodes& cnodes )
{
    if ( pop_transfer == NULL )
    {
        return;
    }
    cube::value_container      inclusive_values;
    cube::value_container      exclusive_values;
    cube::list_of_sysresources lsysres;
    cube::IdIndexMap           metric_id_indices;
    cube->getMetricSubtreeValues( cnodes,
                                  lsysres,
                                  *pop_transfer,
                                  0,
                                  metric_id_indices,
                                  &inclusive_values,
                                  &exclusive_values );
    double pop_transfer_value = inclusive_values[ 0 ]->getDouble();
    setValue( pop_transfer_value );
}


const std::string&
POPTransferTest::getCommentText() const
{
    if ( pop_transfer == NULL )
    {
        return no_comment;
    }
    else
    {
        return no_comment;
    }
}


// ------ overview tests ---------

bool
POPTransferTest::isActive() const
{
    return pop_transfer != NULL;
};

bool
POPTransferTest::isIssue() const
{
    return false;
};

void
POPTransferTest::adjustForTest( cube::CubeProxy* cube ) const
{
    if ( scout_metrics_available( cube ) )
    {
        cube::Metric* max_total_time_ideal = cube->getMetric( "max_total_time_ideal" );
        if ( max_total_time_ideal == NULL )
        {
            add_max_total_time_ideal( cube );
        }
        cube::Metric* max_total_time = cube->getMetric( "max_total_time" );
        if ( max_total_time_ideal == NULL )
        {
            add_max_total_time( cube );
        }
        max_total_time_ideal = cube->getMetric( "max_total_time_ideal" );
        max_total_time       = cube->getMetric( "max_total_time" );
        if ( max_total_time_ideal != NULL && max_total_time != NULL )
        {
            add_transfer_eff( cube );
        }
    }
}


void
POPTransferTest::add_transfer_eff( cube::CubeProxy* ) const
{
    cube::Metric* _met = cube->defineMetric(
        tr( "Transfer efficiency" ).toUtf8().data(),
        "transf_eff",
        "DOUBLE",
        "",
        "",
        POP_TRANSFER_EFF_METRIC_URL,
        tr( "Transfer efficiency reflects the loss of efficiency due to actual data transfer, i.e. ( max_total_time_ideal / max_total_time ). The value varies from 0 to 100%, where 100% shows that routine is a pure computation routine and if MPI does not include waiting time. Do not consider values in the flat view and the system tree." ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_POSTDERIVED,
        "metric::max_total_time_ideal() / metric::max_total_time()",
        "",
        "",
        "",
        "",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}

void
POPTransferTest::add_max_total_time_ideal( cube::CubeProxy* ) const
{
    cube::Metric* _comp = cube->getMetric( "execution" );
    if ( _comp == NULL )
    {
        add_execution_time( cube );
    }
    cube::Metric* _transfer_time_mpi = cube->getMetric( "transfer_time_mpi" );
    if ( _transfer_time_mpi == NULL )
    {
        return;
    }

    cube::Metric* _met = cube->defineMetric(
        tr( "Maximal total time in ideal network" ).toUtf8().data(),
        "max_total_time_ideal",
        "DOUBLE",
        tr( "sec" ).toUtf8().data(),
        "",
        POP_TRANSFER_EFF_METRIC_URL,
        tr( "Maximal total time in ideal network, ( execution - transfer_time_mpi )" ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_PREDERIVED_INCLUSIVE,
        "metric::execution() - metric::transfer_time_mpi()",
        "",
        "",
        "",
        "max(arg1, arg2)",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}


void
POPTransferTest::add_max_total_time( cube::CubeProxy* ) const
{
    cube::Metric* _comp = cube->getMetric( "execution" );
    if ( _comp == NULL )
    {
        add_execution_time( cube );
    }
    cube::Metric* _met = cube->defineMetric(
        tr( "Maximal total time" ).toUtf8().data(),
        "max_total_time",
        "DOUBLE",
        tr( "sec" ).toUtf8().data(),
        "",
        POP_TRANSFER_EFF_METRIC_URL,
        tr( "Maximal total time, max( execution )" ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_PREDERIVED_INCLUSIVE,
        "metric::execution()",
        "",
        "",
        "",
        "max(arg1, arg2)",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}
